from AccessControl import ClassSecurityInfo
from Acquisition import Implicit
from App.class_init import InitializeClass
from OFS.SimpleItem import Item
from Persistence import Persistent
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.permissions import ManageProperties
from Products.CMFCore.utils import getToolByName
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PortalTransforms.interfaces import IChain
from Products.PortalTransforms.interfaces import ITransform
from Products.PortalTransforms.utils import _www
from UserList import UserList
from zope.interface import implementer


try:
    from OFS.role import RoleManager
except ImportError:
    # Zope <=2.12
    from AccessControl.Role import RoleManager


@implementer(IChain, ITransform)
class chain(UserList):
    """A chain of transforms used to transform data"""

    def __init__(self, name='', *args):
        UserList.__init__(self, *args)
        self.__name__ = name
        if args:
            self._update()

    def name(self):
        return self.__name__

    def registerTransform(self, transform):
        self.append(transform)

    def unregisterTransform(self, name):
        for i in range(len(self)):
            tr = self[i]
            if tr.name() == name:
                self.pop(i)
                break
        else:
            raise Exception('No transform named %s registered' % name)

    def convert(self, orig, data, **kwargs):
        for transform in self:
            data = transform.convert(orig, data, **kwargs)
            orig = data.getData()
        md = data.getMetadata()
        md['mimetype'] = self.output
        return data

    def __setitem__(self, key, value):
        UserList.__setitem__(self, key, value)
        self._update()

    def append(self, value):
        UserList.append(self, value)
        self._update()

    def insert(self, *args):
        UserList.insert(*args)
        self._update()

    def remove(self, *args):
        UserList.remove(*args)
        self._update()

    def pop(self, *args):
        UserList.pop(*args)
        self._update()

    def _update(self):
        self.inputs = self[0].inputs
        self.output = self[-1].output
        for i in range(len(self)):
            if hasattr(self[-i - 1], 'output_encoding'):
                self.output_encoding = self[-i - 1].output_encoding
                break
        else:
            try:
                del self.output_encoding
            except:
                pass


class TransformsChain(Implicit, Item, RoleManager, Persistent):
    """ a transforms chain is suite of transforms to apply in order.
    It follows the transform API so that a chain is itself a transform.
    """

    module = 'N/A'
    meta_type = 'TransformsChain'

    meta_types = all_meta_types = ()

    manage_options = (
        ({'label': 'Configure',
          'action': 'manage_main'},
         {'label': 'Reload',
          'action': 'manage_reloadTransform'},) +
        Item.manage_options)

    manage_main = PageTemplateFile('editTransformsChain', _www)
    manage_reloadTransform = PageTemplateFile('reloadTransform', _www)

    security = ClassSecurityInfo()

    def __init__(self, id, description, ids=()):
        self.id = id
        self.description = description
        self._object_ids = list(ids)
        self.inputs = ('application/octet-stream',)
        self.output = 'application/octet-stream'
        self._chain = None

    def _chain_init(self):
        """ build the transforms chain """
        tr_tool = getToolByName(self, 'portal_transforms')
        self._chain = c = chain()
        for id in self._object_ids:
            object = getattr(tr_tool, id)
            c.registerTransform(object)
        self.inputs = c.inputs or ('application/octet-stream',)
        self.output = c.output or 'application/octet-stream'

    @security.public
    def convert(self, *args, **kwargs):
        """ return apply the transform and return the result """
        if self._chain is None:
            self._chain_init()
        return self._chain.convert(*args, **kwargs)

    @security.public
    def name(self):
        """return the name of the transform instance"""
        return self.id

    @security.private
    def manage_beforeDelete(self, item, container):
        Item.manage_beforeDelete(self, item, container)
        if self is item:
            # unregister self from catalog on deletion
            tr_tool = getToolByName(self, 'portal_transforms')
            tr_tool.unregisterTransform(self.id)

    @security.protected(ManagePortal)
    def manage_addObject(self, id, REQUEST=None):
        """ add a new transform or chain to the chain """
        assert id not in self._object_ids
        self._object_ids.append(id)
        self._chain_init()
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main')

    @security.protected(ManagePortal)
    def manage_delObjects(self, ids, REQUEST=None):
        """ delete the selected mime types """
        for id in ids:
            self._object_ids.remove(id)
        self._chain_init()
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main')

    # transforms order handling #

    @security.protected(ManagePortal)
    def move_object_to_position(self, id, newpos):
        """ overriden from OrderedFolder to store id instead of objects
        """
        oldpos = self._object_ids.index(id)
        if (newpos < 0 or newpos == oldpos or newpos >= len(self._object_ids)):
            return 0
        self._object_ids.pop(oldpos)
        self._object_ids.insert(newpos, id)
        self._chain_init()
        return 1

    @security.protected(ManageProperties)
    def move_object_up(self, id, REQUEST=None):
        """  move object with the given id up in the list """
        newpos = self._object_ids.index(id) - 1
        self.move_object_to_position(id, newpos)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main')

    @security.protected(ManageProperties)
    def move_object_down(self, id, REQUEST=None):
        """  move object with the given id down in the list """
        newpos = self._object_ids.index(id) + 1
        self.move_object_to_position(id, newpos)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main')

    # Z transform interface #

    @security.protected(ManagePortal)
    def reload(self):
        """ reload the module where the transformation class is defined """
        for tr in self.objectValues():
            tr.reload()
        self._chain_init()

    # utilities #

    @security.protected(ManagePortal)
    def listAddableObjectIds(self):
        """ return a list of addable transform """
        tr_tool = getToolByName(self, 'portal_transforms')
        return [id for id in tr_tool.objectIds()
                if not (id == self.id or id in self._object_ids)]

    @security.protected(ManagePortal)
    def objectIds(self):
        """ return a list of addable transform """
        return tuple(self._object_ids)

    @security.protected(ManagePortal)
    def objectValues(self):
        """ return a list of addable transform """
        tr_tool = getToolByName(self, 'portal_transforms')
        return [getattr(tr_tool, id) for id in self.objectIds()]

InitializeClass(TransformsChain)



# -*- coding: utf-8 -*-
"""
Original code from active state recipe
        'Colorize Python source using the built-in tokenizer'

----------------------------------------------------------------------------
     MoinMoin - Python Source Parser

 This code is part of MoinMoin (http://moin.sourceforge.net/) and converts
 Python source code to HTML markup, rendering comments, keywords, operators,
 numeric and string literals in different colors.

 It shows how to use the built-in keyword, token and tokenize modules
 to scan Python source code and re-emit it with no changes to its
 original formatting (which is the hard part).
"""

from cStringIO import StringIO
from DocumentTemplate.DT_Util import html_quote
from Products.PortalTransforms.interfaces import ITransform
from zope.interface import implementer

import keyword
import string
import token
import tokenize


# Python Source Parser #####################################################

_KEYWORD = token.NT_OFFSET + 1
_TEXT = token.NT_OFFSET + 2


class Parser:
    """ Send colored python source.
    """

    def __init__(self, raw, tags, out):
        """ Store the source text.
        """
        self.raw = string.strip(string.expandtabs(raw))
        self.out = out
        self.tags = tags

    def format(self):
        """ Parse and send the colored source.
        """
        # store line offsets in self.lines
        self.lines = [0, 0]
        pos = 0
        while True:
            pos = string.find(self.raw, '\n', pos) + 1
            if not pos:
                break
            self.lines.append(pos)
        self.lines.append(len(self.raw))

        # parse the source and write it
        self.pos = 0
        text = StringIO(self.raw)
        self.out.write('<pre class="python">\n')
        try:
            tokenize.tokenize(text.readline, self)
        except tokenize.TokenError as ex:
            msg = ex[0]
            line = ex[1][0]
            self.out.write("<h5 class='error>'ERROR: %s%s</h5>" % (
                msg, self.raw[self.lines[line]:]))
        self.out.write('\n</pre>\n')

    def __call__(self, toktype, toktext, sx, ex, line):
        """ Token handler.
        """
        (srow, scol) = sx
        (erow, ecol) = ex
        oldpos = self.pos
        newpos = self.lines[srow] + scol
        self.pos = newpos + len(toktext)

        # handle newlines
        if toktype in [token.NEWLINE, tokenize.NL]:
            self.out.write('\n')
            return

        # send the original whitespace, if needed
        if newpos > oldpos:
            self.out.write(self.raw[oldpos:newpos])

        # skip indenting tokens
        if toktype in [token.INDENT, token.DEDENT]:
            self.pos = newpos
            return

        # map token type to a group
        if token.LPAR <= toktype and toktype <= token.OP:
            toktype = 'OP'
        elif toktype == token.NAME and keyword.iskeyword(toktext):
            toktype = 'KEYWORD'
        else:
            toktype = tokenize.tok_name[toktype]

        open_tag = self.tags.get('OPEN_' + toktype, self.tags['OPEN_TEXT'])
        close_tag = self.tags.get('CLOSE_' + toktype, self.tags['CLOSE_TEXT'])

        # send text
        self.out.write(open_tag)
        self.out.write(html_quote(toktext))
        self.out.write(close_tag)


@implementer(ITransform)
class PythonTransform(object):
    """Colorize Python source files
    """

    __name__ = "python_to_html"
    inputs = ("text/x-python",)
    output = "text/html"

    config = {
        'OPEN_NUMBER': '<span style="color: #0080C0;">',
        'CLOSE_NUMBER': '</span>',
        'OPEN_OP': '<span style="color: #0000C0;">',
        'CLOSE_OP': '</span>',
        'OPEN_STRING': '<span style="color: #004080;">',
        'CLOSE_STRING': '</span>',
        'OPEN_COMMENT': '<span style="color: #008000;">',
        'CLOSE_COMMENT': '</span>',
        'OPEN_NAME': '<span style="color: #000000;">',
        'CLOSE_NAME': '</span>',
        'OPEN_ERRORTOKEN': '<span style="color: #FF8080;">',
        'CLOSE_ERRORTOKEN': '</span>',
        'OPEN_KEYWORD': '<span style="color: #C00000;">',
        'CLOSE_KEYWORD': '</span>',
        'OPEN_TEXT': '',
        'CLOSE_TEXT': '',
    }

    def name(self):
        return self.__name__

    def convert(self, orig, data, **kwargs):
        dest = StringIO()
        Parser(orig, self.config, dest).format()
        data.setData(dest.getvalue())
        return data


def register():
    return PythonTransform()
